

package name.panitz.util;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Consumer;
import java.util.Comparator;

import java.util.NoSuchElementException;

public sealed interface LL<E> permits LL.Nil, LL.Cons{

  static int binSearch(int x,int[] xs){
    var low = 0;
    var up  = xs.length-1;
    while (low<=up){
      var middle = (low+up)/2;
      if (xs[middle]==x) return middle;
      if (low==middle) low=up;
      if (low==up) break;
      if (xs[middle]>x) up = middle;
      else if (xs[middle]<x) low = middle;
    }
    return -1;
  }
  
  record Nil<E>() implements LL<E>{
    @Override public String toString(){return "[]";}
  }

  record Cons<E>(E hd,LL<E> tl) implements LL<E>{
    @Override public String toString(){
      return show(true,new StringBuffer("["));
    } 
  }

  default boolean isEmpty(){return this instanceof Nil;}

  default E head(){
    if (this instanceof Cons<E> c) return c.hd();
    throw new NoSuchElementException("head on empty list");
  }

  default LL<E> tail(){
    if (this instanceof Cons<E> c) return c.tl();
    throw new NoSuchElementException("tail on empty list");
  }

  @SuppressWarnings("rawtypes")
  final LL nil = new Nil<>();
  @SuppressWarnings("unchecked")
  static <E>LL<E> nil(){return nil;}

  static <E>LL<E> cons(E hd,LL<E> tl){return new Cons<>(hd,tl);}

  @SafeVarargs
  static <E>LL<E> of(E...es){
    LL<E> r = nil();
    for (int i=es.length-1;i>=0;i--) r = cons(es[i],r);
    return r; 
  }

  default String show(boolean first,StringBuffer result){
    if (isEmpty()) {
      result.append("]");
      return result.toString();
    }
    if (!first)result.append(", ");
    result.append(head());
    return tail().show(false,result);
  }


  default int length(){
    if (isEmpty())return 0;
    return 1+tail().length();
    /*var r = 0;
    for (var it=this;!it.isEmpty();it=it.tail())
      r++;
      return r;*/

  }


  default E last1(){
    if (tail().isEmpty()) return head();
    return tail().last1();
  }

  /* Version mit Deconstruction Pattern. Seit Java 19 ein neues preview feature.
     (scheint ab Java 21 festes feature zu werden.)
   */
  @SuppressWarnings("preview")
  default E last(){
    return switch (this) {
    case Cons<E>(E x,Nil<E>()) -> x;
    case Cons<E> xs -> xs.tail().last();
    case Nil<E>() -> throw new NoSuchElementException();
    };  
  }




  default LL<E> append(LL<E> that){
    if (this instanceof Cons) return cons(head(),tail().append(that)); 
    else return that;
  }




  default LL<E> drop(int i){
    if (this instanceof Cons<E> c && i>0)
      return c.tail().drop(i-1);
    return this;
  }



    
  default LL<E> take(int i){
    if (this instanceof Cons && i>0)
      return cons(head(),tail().take(i-1));
    return nil();
  }




  default LL<E> sublist(int from, int length) {
    return drop(from).take(length);
  }



  /* Die ineffizientere rekursive Lösung. */
  default LL<E> reverse1(){
    if (this instanceof Cons)
      return tail().reverse1().append(cons(head(),nil()));
    return nil();
  }

  /* Die iterative Lösung. */
  default LL<E> reverse(){
    LL<E> rs = nil();
    for (var it=this;!it.isEmpty();it=it.tail())
      rs = cons(it.head(),rs);
    return rs;
  }



  default LL<E> intersperse(E e){
    if (isEmpty()||tail().isEmpty()) return this;
    return cons(head(),cons(e,tail().intersperse(e)));
  }



  default boolean isPrefixOf(LL<E> that){
    if (isEmpty()) return true;
    if (that.isEmpty()) return false;
    return head().equals(that.head())&&tail().isPrefixOf(that.tail());
  }



  default boolean isSuffixOf(LL<E> that){
    return this.reverse().isPrefixOf(that.reverse());
  }



  default boolean isInfixOf(LL<E> that){
    return isPrefixOf(that)||
      (!that.isEmpty()&&isInfixOf(that.tail()));
  }



  default E get(int i){
    if (isEmpty()||i<0) throw new IndexOutOfBoundsException();
    if (i==0) return head();
    return tail().get(i-1);
  }      



  default LL<E> rotate(){return tail().append(of(head()));}



  default LL<LL<E>> tails(){
    if (isEmpty()) return of(nil());
    return cons(this,tail().tails());
  }


  
  default void forEach(Consumer<? super E> con) {
    if (isEmpty()) return;
    con.accept(head());
    tail().forEach(con);
  }




  default boolean containsWith(Predicate< ? super E> p) {
    if (isEmpty()) return false;
    return p.test(head())||tail().containsWith(p);
  }




  default boolean contains(E el) {
    return containsWith(x->x.equals(el));
  }
  



  default LL<E> dropWhile(Predicate< ? super E> p){
    if (this instanceof Cons<E> c && p.test(head()))
      return c.tail().dropWhile(p);
    return this;
  }



  default LL<E> takeWhile(Predicate< ? super E> p){
    if (this instanceof Cons && p.test(head()))
      return cons(head(),tail().takeWhile(p));
    return nil();
  }



  default LL<E> filter(Predicate<? super E> p){
    if (isEmpty()) return nil();
    if (p.test(head())) return cons(head(), tail().filter(p));
    return tail().filter(p);
  }



  default <R> LL<R> map(Function<? super E, ? extends R> f){
    if (this instanceof Cons)
      return cons(f.apply(head()), tail().map(f));
    return nil();    
  }


  record Pair<A,B>(A fst,B snd){
    @Override public String toString(){return "("+fst()+", "+snd()+")";}
  }    


  default <B> LL<Pair<E,B>> zip(LL<B> that){
    if (isEmpty()||that.isEmpty()) return nil();
    return cons(new Pair<>(head(),that.head()),tail().zip(that.tail()));
  }



  default Pair<LL<E>,LL<E>> span(Predicate<? super E> p){
    if (isEmpty()) return new Pair<>(nil(),nil()); 
    if (!p.test(head())) return new Pair<>(nil(),this);
    var r = tail().span(p);
    return new Pair<>(cons(head(),r.fst()),r.snd());
  }





  default Pair<LL<E>,LL<E>> partition(Predicate<? super E> p){
    if (isEmpty()) return new Pair<>(nil(),nil()); 
    var r = tail().partition(p);
    return p.test(head())
      ?new Pair<>(cons(head(),r.fst()),r.snd())
      :new Pair<>(r.fst(),cons(head(),r.snd()))
    ;
  }





  default boolean isSorted(Comparator<? super E> cmp){
    return isEmpty()
     ||tail().isEmpty()
     ||cmp.compare(head(),tail().head())<=0 && tail().isSorted(cmp);
  }





  default LL<E> qsort(Comparator<? super E> cmp){
    if (isEmpty()) return nil();
    var p = tail().partition(x->cmp.compare(head(),x)>=0);
    return p.fst().qsort(cmp).append(cons(head(),p.snd().qsort(cmp)));
  }




}
